import logging

__all__ = ("logger", "init_logging")

logger = logging.getLogger("track-bump")


def init_logging(level: int = logging.WARNING):
    stream_handler = logging.StreamHandler()
    formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    logger.setLevel(level)



from .utils import get_last_tag, parse_version, get_last_commit_message
from .logs import logger
from .env import DEFAULT_BRANCH
from typing import Literal
import re

__all__ = (
    "get_latest_default_tag",
    "get_latest_release_tag",
    "get_branch_release_tag",
    "get_new_tag",
    "ReleaseTag",
    "BranchName",
)

ReleaseTag = Literal["beta", "rc", "stable"]
type BranchName = str

_RELEASE_TAGS: dict[BranchName, ReleaseTag] = {
    r"^develop$": "beta",
    r"^release/.*": "rc",
    rf"^{DEFAULT_BRANCH}$": "stable",
}


def get_latest_default_tag():
    return get_last_tag(r"^v\d+\.\d+\.\d+$")


def get_latest_release_tag(release_tag: str) -> str | None:
    return get_last_tag(rf"^v\d+\.\d+\.\d+-{release_tag}\.\d+$")


def get_branch_release_tag(branch: BranchName) -> ReleaseTag:
    for branch_pattern, release_tag in _RELEASE_TAGS.items():
        if re.match(branch_pattern, branch):
            return release_tag
    raise ValueError(f"Branch {branch!r} is not supported")


_BUMP_MINOR_REG = re.compile(r"release:.*")


def get_new_tag(latest_tag: str | None, release_tag: ReleaseTag, last_commit_message: str | None = None) -> str:
    """
    Return the new tag based on the latest release tag and current branch
    """
    if not latest_tag:
        raise ValueError("No tags found. Please create a release tag first (like v0.1.0)")
    logger.info(f"Latest tag: {latest_tag}")

    (major, minor, patch), _ = parse_version(latest_tag.removeprefix("v"))
    _next_release = f"v{major}.{minor + 1}.0"
    if release_tag != "stable":
        _latest_release_tag = get_latest_release_tag(release_tag)
        logger.info(f"Latest {release_tag} tag: {_latest_release_tag}")
        _release_number = int(_latest_release_tag.split(".")[-1]) + 1 if _latest_release_tag is not None else 0
        _tag = f"{_next_release}-{release_tag}.{_release_number}"
    else:
        _last_commit_message = last_commit_message or get_last_commit_message()
        logger.info(
            f"Last commit message: {_last_commit_message}"
            if _last_commit_message is not None
            else "No commit message found"
        )
        if _last_commit_message is None or _BUMP_MINOR_REG.match(_last_commit_message):
            logger.debug("Bumping minor")
            _tag = _next_release
        else:
            logger.debug("Bumping patch")
            _tag = f"v{major}.{minor}.{patch + 1}"

    return _tag



from pathlib import Path
import tomllib
import re
from typing import TypedDict

from .logs import logger

__all__ = ("parse_config_file", "replace_in_file", "ParseConfig", "replace_in_files")


def replace_in_file(file_path: Path, version: str, tag: str):
    _new_test = re.sub(rf'^{tag} = "(.*)"$', rf'{tag} = "{version}"', file_path.read_text(), flags=re.MULTILINE)

    file_path.write_text(_new_test)


class ParseConfig(TypedDict):
    version: str
    bump_message: str
    version_files: list[str]


def parse_config_file(config_path: Path) -> ParseConfig:
    if not config_path.exists():
        raise FileNotFoundError(f"{config_path} not found")
    logger.debug(f"Parsing {config_path}")
    data = tomllib.loads(config_path.read_text())

    _config = data.get("tool", {}).get("track-bump")
    if _config is None:
        raise ValueError("Could not find config tool.track-bump in file_path")

    version = _config.get("version")
    if version is None:
        raise ValueError("version is required in config file")
    bump_message = _config.get("bump_message")
    if bump_message is None:
        raise ValueError("bump_message is required in config file")
    version_files = _config.get("version_files", [])
    return {"version": version, "bump_message": bump_message, "version_files": version_files}


def replace_in_files(config_path: Path, files: list[str], version: str):
    replace_in_file(config_path, version=version, tag="version")
    for _file in files:
        try:
            _path, _tag = _file.split(":")
        except ValueError:
            _path = _file
            _tag = "version"
        _file_path = Path(config_path.parent / _path)
        if not _file_path.exists():
            raise FileNotFoundError(f"{_file_path} not found")
        logger.debug(f"Replacing in {_file_path}")
        replace_in_file(_file_path, version=version, tag=_tag)



import logging
from pathlib import Path

from piou import Cli, Option
from .logs import init_logging
from .bump import bump_project
from .env import DEFAULT_BRANCH
from .tags import get_latest_default_tag, get_branch_release_tag, get_latest_release_tag
from .utils import set_cd, get_current_branch

cli = Cli("Track-bump utility commands")

cli.add_option("-v", "--verbose", help="Verbosity")
cli.add_option("-vv", "--verbose2", help="Increased verbosity")


def on_process(verbose: bool = False, verbose2: bool = False):
    init_logging(logging.DEBUG if verbose2 else logging.INFO if verbose else logging.WARNING)


cli.set_options_processor(on_process)


@cli.command(cmd="bump", help="Bump project version")
def bump(
    project_path: Path = Option(Path.cwd(), "-p", "--project", help="Project path"),
    sign_commits: bool = Option(False, "--sign", help="Sign commits"),
    branch: str | None = Option(None, "--branch", help="Branch to bump"),
):
    bump_project(project_path, sign_commits, branch=branch)


@cli.command(cmd="get-latest-tag", help="Get the latest tag")
def get_latest_tag(
    project_path: Path = Option(Path.cwd(), "-p", "--project", help="Project path"),
    branch: str | None = Option(None, "--branch", help="Branch to bump"),
):
    with set_cd(project_path):
        _branch = branch or get_current_branch()
        tag = (
            get_latest_default_tag()
            if _branch == DEFAULT_BRANCH
            else get_latest_release_tag(get_branch_release_tag(_branch))
        )
    if tag:
        print(tag)


def run():
    cli.run()


if __name__ == "__main__":
    run()



from pathlib import Path
from .logs import logger

from track_bump.utils import set_cd, get_current_branch, git_setup, git_commit, git_tag, parse_version
from track_bump.tags import get_latest_default_tag, get_branch_release_tag, get_new_tag
from track_bump.update_files import parse_config_file, replace_in_files

CONFIG_FILES = [".cz.toml", "pyproject.toml"]


def bump_project(
    project_path: Path, sign_commits: bool = False, branch: str | None = None, last_commit_message: str | None = None
):
    for file in CONFIG_FILES:
        config_path = Path(project_path / file)
        if config_path.exists():
            break
    else:
        raise FileNotFoundError(f"Could not find any of the following files: {CONFIG_FILES} in {project_path}")

    config = parse_config_file(config_path)

    git_setup(sign_commits=sign_commits)
    with set_cd(project_path):
        _latest_tag = get_latest_default_tag()
        logger.info(f"Latest tag: {_latest_tag}")
        _branch = branch or get_current_branch()
        _tag = get_branch_release_tag(_branch)
        logger.info(f"Branch {_branch} (tag: {_tag})")

        if _latest_tag is None:
            (major, minor, path), release = parse_version(config["version"])
            _latest_tag = f"v{major}.{max(minor - 1, 1)}.{path}"
        _new_tag = get_new_tag(latest_tag=_latest_tag, release_tag=_tag, last_commit_message=last_commit_message)

    current_version = config["version"]
    new_version = _new_tag.removeprefix("v")
    logger.info(f"New tag {_new_tag} (version: {new_version})")
    replace_in_files(config_path, config["version_files"], new_version)
    _bump_message = config["bump_message"].format(current_version=current_version, new_version=new_version)
    logger.info(f"Committing with message: {_bump_message}")
    with set_cd(project_path):
        git_commit(_bump_message)
        git_tag(_new_tag)
    logger.info("Done")



import os
import pathlib
import re
import subprocess
from track_bump.env import CI_USER, CI_USER_EMAIL
import contextlib
from .logs import logger

__all__ = (
    "exec_cmd",
    "get_last_tag",
    "git_tag",
    "git_setup",
    "set_cd",
    "get_current_branch",
    "git_commit",
    "parse_version",
    "get_tags",
    "get_last_commit_message",
)


def exec_cmd(cmd: str | list[str], *, encoding: str = "utf-8", env: dict | None = None) -> str:
    default_shell = os.getenv("SHELL", "/bin/bash")

    logger.debug(f"Executing command: {cmd}")
    stdout, stderr = subprocess.Popen(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, executable=default_shell, env=env
    ).communicate()
    if stderr:
        raise Exception(stderr.decode(encoding))
    return stdout.decode(encoding)


@contextlib.contextmanager
def set_cd(path: pathlib.Path):
    prev_cwd = pathlib.Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def get_tags():
    tags = exec_cmd("git tag --sort=-version:refname").split("\n")
    return [x.strip() for x in tags if x.strip()]


def get_last_tag(pattern: str) -> str | None:
    _tags = get_tags()
    _valid_tags = [_tag for _tag in _tags if re.match(pattern, _tag)]
    return _valid_tags[0] if _valid_tags else None


def git_tag(version: str):
    _output = exec_cmd(f"git tag {version}")
    if _output:
        logger.debug(f"tag output: {_output}")


def git_setup(sign_commits: bool = False):
    if CI_USER is None:
        raise ValueError("CI_USER must be set")
    if CI_USER_EMAIL is None:
        raise ValueError("CI_USER_EMAIL must be set")

    exec_cmd(f'git config user.email "{CI_USER_EMAIL}"')
    exec_cmd(f'git config user.name "{CI_USER}"')
    if sign_commits:
        exec_cmd("git config commit.gpgSign true")


def get_current_branch() -> str:
    return exec_cmd("git branch --show-current").strip()


def git_commit(message: str):
    exec_cmd("git add .")
    exec_cmd(f'git commit -am "{message}"')


def get_last_commit_message() -> str | None:
    _latest_commit = exec_cmd("git log -1 --pretty=%B").strip()
    return _latest_commit if _latest_commit else None


type MajorMinorPatch = tuple[int, int, int]
type ReleaseVersion = tuple[str, int]


def parse_version(version: str) -> tuple[MajorMinorPatch, ReleaseVersion | None]:
    _version, *_release = version.split("-")
    major, minor, patch = [int(x) for x in _version.split(".")]

    if _release:
        _release_name, _release_number_str = _release[0].split(".")
        release = (_release_name, int(_release_number_str))
    else:
        release = None
    return (major, minor, patch), release
